package users

import (
	"encoding/json"
	"errors"
	"os"
	"sort"
	"sync"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/config"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/logger"
)

type Users struct {
	m      map[string]User
	file   string
	rwlock *sync.RWMutex
}

var log = logger.GetInstance()
var conf = config.GetInstance()
var users *Users

// Returns a Users "singleton".
// IMPORTANT: the InitUsers function must have been previously called!
func GetUsersInstance() *Users {
	return users
}

// Creates users by reading them from the config directory.
func InitUsers() error {
	usersFile := conf.UsersFile
	users = &Users{m: make(map[string]User), file: usersFile, rwlock: new(sync.RWMutex)}
	return users.createUsersFromFile()
}

// Returns all users.
func (users *Users) GetUsers() []User {
	users.rwlock.RLock()
	list := users.usersToList()
	users.rwlock.RUnlock()

	// Sort users by names
	sort.Slice(list, func(i, j int) bool {
		return list[i].LastName < list[j].LastName
	})
	return list
}

// Returns a user by her email (unique).
func (users *Users) GetUserByEmail(email string) (User, error) {
	users.rwlock.RLock()
	defer users.rwlock.RUnlock()

	for _, u := range users.m {
		if email == u.Email {
			return u, nil
		}
	}

	return dummyUser, errors.New(email + ": user not found")
}

// Deletes a user by its email and upates the user file.
// REMARKS:
// After a user is deleted, their VMs and templates remain.
// Alternative: make sure a user's VMs and templates are first deleted before deleting a user.
// I think the current behavior is preferable.
func (users *Users) DeleteUserByEmail(email string) error {
	users.rwlock.Lock()
	defer users.rwlock.Unlock()
	key := email
	_, exists := users.m[key]
	if exists {
		delete(users.m, key)
		err := users.writeUsersFile()
		return err
	}
	return errors.New(email + ": user not found")
}

// Adds a user and updates the users file.
func (users *Users) AddUser(user *User) error {
	users.rwlock.Lock()
	defer users.rwlock.Unlock()

	// Checks the username doesn't already exists.
	for _, u := range users.m {
		if user.Email == u.Email {
			return errors.New("User already exists")
		}
	}

	key := user.Email
	users.m[key] = *user
	err := users.writeUsersFile()
	return err
}

// Updates a user and updates the user file.
func (users *Users) UpdateUser(user *User) error {
	users.rwlock.Lock()
	defer users.rwlock.Unlock()
	key := user.Email
	_, exists := users.m[key]
	if exists {
		// Deletes previous user and adds new one
		delete(users.m, key)
		users.m[key] = *user
		err := users.writeUsersFile()
		return err
	}
	return errors.New(key + ": user not found")
}

// Creates users by reading them from the user file.
func (users *Users) createUsersFromFile() error {
	filein, err := os.OpenFile(users.file, os.O_RDONLY, 0)
	if err != nil {
		return errors.New("Failed reading users file: " + err.Error())
	}
	defer filein.Close()

	// Decodes the json file into users, then checks:
	// - all fields are present
	// - extra fields are not allowed
	decoder := json.NewDecoder(filein)
	decoder.DisallowUnknownFields()
	usersAsList := []User{}
	if err = decoder.Decode(&usersAsList); err != nil {
		return errors.New("Failed decoding users file \"" + users.file + "\": " + err.Error())
	}

	// Validates each user entry
	for _, u := range usersAsList {
		if err = u.Validate(); err != nil {
			return errors.New("Failed validating users file \"" + users.file + "\": " + err.Error())
		}
		// Add the entry to the map of users
		key := u.Email
		users.m[key] = u
	}

	return nil
}

// Writes users to file.
func (users *Users) writeUsersFile() error {
	file, err := os.Create(users.file)
	if err != nil {
		log.Error("Failed writing users file: " + err.Error())
		return errors.New("Failed writing users file: " + err.Error())
	}
	defer file.Close()

	encoder := json.NewEncoder(file)
	encoder.SetIndent("", "    ")
	usersAsList := users.usersToList()
	if err = encoder.Encode(&usersAsList); err != nil {
		log.Error("Failed encoding users file \"" + users.file + "\": " + err.Error())
		return errors.New("Failed encoding users file \"" + users.file + "\": " + err.Error())
	}
	return nil
}

// Converts the map of users into a list.
func (users *Users) usersToList() []User {
	list := []User{}
	for _, val := range users.m {
		list = append(list, val)
	}
	return list
}
